/*
    Copyright (C) 1999 by Joe Pasqua <joe@NoRoomAtTheInn.org>.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.


    THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
    OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
    OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

import java.util.Hashtable;
import java.util.Vector;

/**
CPUXA {
    boolean receiveX10(X10Code code);
    boolean x10State(X10Code code, NumberInt(0-5) state);

    void xmitX10(X10Code code);
    void xmitX10Rpt(X10Code code, int(0-15) repeat);
    void x10QuickOn(X10Code);
    void x10QuickOff(X10Code);

    void pcsPresetDim(X10Code code, int(0-31) dimLevel);
    void levitonPresetDim(X10Code code, int(0-63) dimLevel);
    void pcsPresetDim%(X10Code code, int(0-100) dimLevel);
    void levitonPresetDim%(X10Code code, int(0-100) dimLevel);
    void levitonGroupOn(int(0-15) houseCode)
    void levitonGroupPreset(int(0-15) houseCode)

    void xmitIR(int(1-511) code);
    boolean recvIR(int(1-511) code);

    void sendPage(int(1-15) msgNum);
    void sendAscii(int(0-127) stringNum);

    boolean ioError();
}
*/

class CPUXAType extends Type {

    CPUXAType() {
        methods = new Hashtable(16);
        methods.put("recvX10", new RecvX10Method());
        methods.put("receiveIR", new RecvIRMethod());
        methods.put("x10State", new X10StateMethod());
        methods.put("ioError", new IOErrorMethod());
        methods.put("xmitX10", new XmitX10Method(false));
        methods.put("xmitX10Rpt", new XmitX10Method(true));
        methods.put("xmitIR", new XmitIRMethod());
        methods.put("pcsPreset", new PCSPresetDimMethod(false));
        methods.put("pcsPresetPct", new PCSPresetDimMethod(true));
        methods.put("x10QuickOn", new X10QuickMethod(true));
        methods.put("x10QuickOff", new X10QuickMethod(false));
        methods.put("sendPage", new SendPageMethod());
        methods.put("levPreset", new LevitonPresetDimMethod(false));
        methods.put("levPresetPct", new LevitonPresetDimMethod(true));
        methods.put("levGrpOff", new LevitonGroupMethod(false));
        methods.put("levGrpPreset", new LevitonGroupMethod(true));
        methods.put("sendAscii", new SendAsciiMethod());
    }


    class IOErrorMethod implements Type.Method {

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg.getArgs(null, inputs);  // Make sure there are no args
            return Type.encode(0x02, 0x40, 0, 0);
        }
    }

    class RecvIRMethod implements Type.Method {

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("Number", "irCode", "1", "1024") };
            Vector  args = Arg.getArgs(argDesc, inputs);

            return Type.encode(
                    0x01, 0xA0, 0, ((NumberVal)args.elementAt(0)).intValue());
        }
    }

    class XmitIRMethod implements Type.Method {

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("Number", "irCode", "1", "1024") };
            Vector  args = Arg.getArgs(argDesc, inputs);

            return Type.encode(
                0x05, 0x80, 0, ((NumberVal)args.elementAt(0)).intValue());
        }
    }

    class SendAsciiMethod implements Type.Method {

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("Number", "asciiMsgNum", "0", "127") };
            Vector  args = Arg.getArgs(argDesc, inputs);

            return Type.encode(
                0x07, 0x00, 0, ((NumberVal)args.elementAt(0)).intValue());
        }
    }

    class SendPageMethod implements Type.Method {

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("Number", "pageMsgNum", "1", "15") };
            Vector  args = Arg.getArgs(argDesc, inputs);

            return Type.encode(
                0x06, 0x60, ((NumberVal)args.elementAt(0)).intValue(), 0);
        }
    }

    class RecvX10Method implements Type.Method {

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("X10Code", "x10Code") };
            Vector  args = Arg.getArgs(argDesc, inputs);

            return Type.encode(
                0x00, 0x40, ((X10CodeVal)args.elementAt(0)).getEncoded(), 0);
        }
    }

    class XmitX10Method implements Type.Method {
        private boolean repeat;

        XmitX10Method(boolean repeat) { this.repeat = repeat; }

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("X10Code", "x10Code") };
            Arg[]  argDescRpt = {
                    new Arg("X10Code", "x10Code"),
                    new Arg("Number", "repeat", "1", "15")
            };
            Vector  args = Arg.getArgs(repeat ? argDescRpt : argDesc, inputs);

            X10CodeVal  code = (X10CodeVal)args.elementAt(0);

            int addToCode = 0;
            if (repeat) {
                int keyCode = code.getKeyCode();
                if (keyCode != 20 && keyCode != 21) // Not Dim or Bright
                    throw new ParseException(
                            "Only Dim and Bright can be repeated");
                addToCode = ((((NumberVal)args.elementAt(1)).intValue())-1) << 12;
            }

            return Type.encode(
                    0x04, 0x60, code.getEncoded() + addToCode, 0);
        }
    }

    class X10QuickMethod implements Type.Method {

        private boolean turnOn;

        X10QuickMethod(boolean turnOn) { this.turnOn = turnOn; }

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("X10Code", "x10Code") };
            Vector  args = Arg.getArgs(argDesc, inputs);

            return Type.encode(
                0x06, (turnOn ? 0x00 : 0x20),
                ((X10CodeVal)args.elementAt(0)).getEncoded(), 0);
        }
    }

    class PCSPresetDimMethod implements Type.Method {
        private boolean byPct;

        PCSPresetDimMethod(boolean byPct) { this.byPct = byPct; }

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]   argDesc = new Arg[2];
            argDesc[0] = new Arg("X10Code", "x10Code");
            argDesc[1] = byPct ? new Arg("Number", "percent", "0", "100") :
                                new Arg("Number", "dimLevel", "0", "31");
            Vector  args = Arg.getArgs(argDesc, inputs);

            int dimLevel = ((NumberVal)args.elementAt(1)).intValue();
            dimLevel = byPct ? (dimLevel * 31)/100 : dimLevel;

            return Type.encode(
                0x05, 0xE0,
                ((X10CodeVal)args.elementAt(0)).getEncoded(), dimLevel);
        }
    }

    class LevitonPresetDimMethod implements Type.Method {
        private boolean byPct;

        LevitonPresetDimMethod(boolean byPct) {
            this.byPct = byPct; }

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]   argDesc = new Arg[2];
            argDesc[0] = new Arg("X10Code", "x10Code");
            argDesc[1] = byPct ? new Arg("Number", "percent", "0", "100") :
                                new Arg("Number", "dimLevel", "0", "63");
            Vector  args = Arg.getArgs(argDesc, inputs);

            int dimLevel = ((NumberVal)args.elementAt(1)).intValue();
            dimLevel = byPct ? (dimLevel * 63)/100 : dimLevel;

            return Type.encode(
                0x06, 0xA0,
                ((X10CodeVal)args.elementAt(0)).getEncoded(), dimLevel);
        }
    }

    class LevitonGroupMethod implements Type.Method {
        private boolean preset;

        LevitonGroupMethod(boolean preset) { this.preset = preset; }

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = { new Arg("X10Code", "x10Code") };
            Arg[]  argDescPre = {
                    new Arg("X10Code", "x10Code"),
                    new Arg("Number", "dimLevel", "0", "63")
            };
            Vector  args = Arg.getArgs(preset ? argDescPre : argDesc, inputs);

            X10CodeVal  code = (X10CodeVal)args.elementAt(0);
            int level = preset ? ((NumberVal)args.elementAt(1)).intValue() : 0;

            return Type.encode(
                0x06, 0xC0,
                ((code.getKeyCode() << 8) + (preset ? 1 : 0)),
                level);
        }
    }

    class X10StateMethod implements Type.Method {

        public short[] codeGen(Value target, Vector inputs)
                throws ParseException
        {
            Arg[]  argDesc = {
                    new Arg("X10Code", "x10Code"),
                    new Arg("Number", "state", "0", "5")
            };
            Vector  args = Arg.getArgs(argDesc, inputs);

            return Type.encode(
                0x01, 0xC0, ((X10CodeVal)args.elementAt(0)).getEncoded(),
                ((NumberVal)args.elementAt(1)).intValue());
        }
    }

}

