/*
    Copyright (C) 1999 by Joe Pasqua <joe@NoRoomAtTheInn.org>.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.


    THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
    OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
    OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

import java.util.Vector;
import java.util.Hashtable;

    /** Base class for all types handled by the compiler. Each actual type
        must subclass this class and must be named typeName + "Type". That is,
        the sublcass representing a type named Number must be called NumberType.
        That's so instances can be created dynamically by name without requiring
        a registration.
    */

abstract class Type {

// ----------------------------------------------------------------------
// Private Class Variables
// ----------------------------------------------------------------------

    /** A table of the types currently know to the compiler. Maps from
        type name to Type. We only create one instance of class Type
        for each type we see. That instance is stored in this table and
        handed out as requested.
    */
    private static Hashtable    types = new Hashtable(10);



// ----------------------------------------------------------------------
// Class Methods
// ----------------------------------------------------------------------

    /** Return an instance of the Type subclass associated with the
        given typeName.
    */
    static Type get(String typeName)
            throws ParseException
    {
        Type    t = (Type)types.get(typeName);

        if (t == null) {
            try {
                t = (Type)(Class.forName(typeName + "Type").newInstance());
                t.typeName = typeName;
                types.put(typeName, t);
            } catch (InstantiationException e) {
                throw new RuntimeException(
                        "Internal Error: Bad type definition of " + typeName);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(
                        "Internal Error: Bad type definition of " + typeName);
            } catch (ClassNotFoundException e) {
                throw new ParseException(
                        "Unknown class in Type.get(): " + typeName);
            }
        }
        return t;
    }

    /** Create an instance of the given type with the given value. The
        returned object will be an instance of a class whose name is
        typeName + "Val". E.g., if typeName is Number, the returned object
        will be of class NumberVal.
    */
    static Value getVal(String typeName, BasicVal value)
            throws ParseException
    {
        return Type.get(typeName)._new((BasicVal)value);
    }

    /** Encode an instruction given the basic elements and return an array
        representing the encoded instruction.
    */
    static short[] encode(
            int opCode, int subCode, int addr, int constant) {
        short[] instr = new short[6];

        instr[0] = (short)subCode;
        instr[1] = (short)opCode;
        instr[2] = (short)(addr & 0x00ff);
        instr[3] = (short)((addr & 0xff00) >> 8);
        instr[4] = (short)(constant & 0x00ff);
        instr[5] = (short)((constant & 0xff00) >> 8);

        return instr;
    }


// ----------------------------------------------------------------------
// Protected Instance Variables
// ----------------------------------------------------------------------

    /** The name of this type; e.g. Number, Timer, Module. This name
        is the root of the actual class name that implements the type.
        The class name is typeName + "Type"; e.g. NumberType or TimerType.
    */
    protected String    typeName;

    /** This table contains the methods implemented by this type and
        is keyed by method name. Each subclass is responsible for filling
        in the table with it's methods.
    */
    protected Hashtable methods;



// ----------------------------------------------------------------------
// Instance Methods
// ----------------------------------------------------------------------


    /** Create a new instance of the this type with the given value (v).
        If isVar is true, this instance represents a CPUXA variable
        (at the moment that is limited to Numbers and Timers). In these
        cases the subclass must "allocate" a CPUXA resource to this instance.
        The returned object is a subclass of Value corresponding to the
        type. E.g., if the type is NumberType the value will be NumberVal.
    */
    Value _new(BasicVal v, boolean isVar)
            throws ParseException
    {
        if (v.isLiteral) {
            String valueClass = typeName + "Val";

            try {
                Value value = (Value)(Class.forName(valueClass).newInstance());
                value.init(this, v.value);
                return value;
            } catch (InstantiationException e) {
                throw new RuntimeException(
                        "Internal Error: Bad type definition of " + typeName);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(
                        "Internal Error: Bad type definition of " + typeName);
            } catch (ClassNotFoundException e) {
                throw new ParseException("Unknown class: " + typeName);
            }
        }
        else {
            Symbol s = XACompiler.symTab.get(v.value);
            if (s == null)
                throw new ParseException("Undefined variable: " + v.value);
            Value theVal = s.getValue();
            if (!theVal.getType().equals(this))
                throw new ParseException(
                    "Type mismatch, found " +
                    theVal.getType() + ", but was expecting " + this);
            if (isVar)
                return theVal.getType()._new(new BasicVal(theVal.toString()), true);
            else
                return theVal;
        }
    }

    /** Shorthand used to create an instance of a non-variable Value.
        This is equivalent to _new(value, false).
    */
    Value _new(BasicVal v) throws ParseException { return _new(v, false); }


    /** Generate code for the spec'd method invoked on the spec'd instance
        with the spec'd argument list. Return an array representing the
        bytes of the resulting instruction.
    */
    short[] codeGen(Value instance, String methodName, Vector args)
            throws ParseException
    {
        Method method = (Method)methods.get(methodName);
        if (method == null) {
            //throw new ParseException("Undefined Method: " + method);
            System.out.println("Method not yet implemented: " + methodName);
            return new short[6];
        }
        return method.codeGen(instance, args);
    }


// ----------------------------------------------------------------------
// Overridden Methods
// ----------------------------------------------------------------------

    public boolean equals(Object otherType) {
        if (otherType == null || !(otherType instanceof Type))
            return false;
        return ((Type)otherType).typeName.equals(typeName);
    }

    public String toString() { return typeName; }



// ----------------------------------------------------------------------
// Nested Classes and Interfcaes
// ----------------------------------------------------------------------

    interface Method {
        abstract short[] codeGen(Value target, Vector args) throws ParseException;
    }


}







